@@ -1,19 +1,22 @@
 from __future__ import annotations
-import hashlib
-from functools import lru_cache, singledispatch
-from inspect import getsource
-import logging
-from parsl.dataflow.taskrecord import TaskRecord
 
-from typing import Dict, Any, List, Optional, TYPE_CHECKING
+import hashlib
+import logging
+import os
+import pickle
+from functools import lru_cache, singledispatch
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence
+
+import typeguard
+
+from parsl.dataflow.errors import BadCheckpoint
+from parsl.dataflow.taskrecord import TaskRecord
 
 if TYPE_CHECKING:
     from parsl import DataFlowKernel  # import loop at runtime - needed for typechecking - TODO turn into "if typing:"
 
-from concurrent.futures import Future
-
-from parsl.serialize import serialize
 import types
+from concurrent.futures import Future
 
 logger = logging.getLogger(__name__)
 
@@ -55,13 +58,13 @@ def id_for_memo(obj: object, output_ref: bool = False) -> bytes:
 @id_for_memo.register(int)
 @id_for_memo.register(float)
 @id_for_memo.register(type(None))
-def id_for_memo_serialize(obj: object, output_ref: bool = False) -> bytes:
-    return serialize(obj)
+def id_for_memo_pickle(obj: object, output_ref: bool = False) -> bytes:
+    return pickle.dumps(obj)
 
 
 @id_for_memo.register(list)
 def id_for_memo_list(denormalized_list: list, output_ref: bool = False) -> bytes:
-    if type(denormalized_list) != list:
+    if type(denormalized_list) is not list:
         raise ValueError("id_for_memo_list cannot work on subclasses of list")
 
     normalized_list = []
@@ -69,12 +72,12 @@ def id_for_memo_list(denormalized_list: list, output_ref: bool = False) -> bytes
     for e in denormalized_list:
         normalized_list.append(id_for_memo(e, output_ref=output_ref))
 
-    return serialize(normalized_list)
+    return pickle.dumps(normalized_list)
 
 
 @id_for_memo.register(tuple)
 def id_for_memo_tuple(denormalized_tuple: tuple, output_ref: bool = False) -> bytes:
-    if type(denormalized_tuple) != tuple:
+    if type(denormalized_tuple) is not tuple:
         raise ValueError("id_for_memo_tuple cannot work on subclasses of tuple")
 
     normalized_list = []
@@ -82,7 +85,7 @@ def id_for_memo_tuple(denormalized_tuple: tuple, output_ref: bool = False) -> by
     for e in denormalized_tuple:
         normalized_list.append(id_for_memo(e, output_ref=output_ref))
 
-    return serialize(normalized_list)
+    return pickle.dumps(normalized_list)
 
 
 @id_for_memo.register(dict)
@@ -92,7 +95,7 @@ def id_for_memo_dict(denormalized_dict: dict, output_ref: bool = False) -> bytes
     When output_ref=True, the values are normalised as output refs, but
     the keys are not.
     """
-    if type(denormalized_dict) != dict:
+    if type(denormalized_dict) is not dict:
         raise ValueError("id_for_memo_dict cannot work on subclasses of dict")
 
     keys = sorted(denormalized_dict)
@@ -101,32 +104,22 @@ def id_for_memo_dict(denormalized_dict: dict, output_ref: bool = False) -> bytes
     for k in keys:
         normalized_list.append(id_for_memo(k))
         normalized_list.append(id_for_memo(denormalized_dict[k], output_ref=output_ref))
-    return serialize(normalized_list)
+    return pickle.dumps(normalized_list)
 
 
 # the LRU cache decorator must be applied closer to the id_for_memo_function call
 # that the .register() call, so that the cache-decorated version is registered.
 @id_for_memo.register(types.FunctionType)
 @lru_cache()
-def id_for_memo_function(function: types.FunctionType, output_ref: bool = False) -> bytes:
-    """This produces function hash material using the source definition of the
-       function.
-
-       The standard serialize_object based approach cannot be used as it is
-       too sensitive to irrelevant facts such as the source line, meaning
-       a whitespace line added at the top of a source file will cause the hash
-       to change.
+def id_for_memo_function(f: types.FunctionType, output_ref: bool = False) -> bytes:
+    """This will checkpoint a function based only on its name and module name.
+    This means that changing source code (other than the function name) will
+    not cause a checkpoint invalidation.
     """
-    logger.debug("serialising id_for_memo_function for function {}, type {}".format(function, type(function)))
-    try:
-        fn_source = getsource(function)
-    except Exception as e:
-        logger.warning("Unable to get source code for app caching. Recommend creating module. Exception was: {}".format(e))
-        fn_source = function.__name__
-    return serialize(fn_source.encode('utf-8'))
+    return pickle.dumps(["types.FunctionType", f.__name__, f.__module__])
 
 
-class Memoizer(object):
+class Memoizer:
     """Memoizer is responsible for ensuring that identical work is not repeated.
 
     When a task is repeated, i.e., the same function is called with the same exact arguments, the
@@ -157,7 +150,7 @@ class Memoizer(object):
 
     """
 
-    def __init__(self, dfk: DataFlowKernel, memoize: bool = True, checkpoint: Dict[str, Future[Any]] = {}):
+    def __init__(self, dfk: DataFlowKernel, *, memoize: bool = True, checkpoint_files: Sequence[str]):
         """Initialize the memoizer.
 
         Args:
@@ -170,6 +163,8 @@ class Memoizer(object):
         self.dfk = dfk
         self.memoize = memoize
 
+        checkpoint = self.load_checkpoints(checkpoint_files)
+
         if self.memoize:
             logger.info("App caching initialized")
             self.memo_lookup_table = checkpoint
@@ -186,9 +181,8 @@ class Memoizer(object):
         Returns:
             - hash (str) : A unique hash string
         """
-        # Function name TODO: Add fn body later
 
-        t = []  # type: List[bytes]
+        t: List[bytes] = []
 
         # if kwargs contains an outputs parameter, that parameter is removed
         # and normalised differently - with output_ref set to True.
@@ -198,23 +192,20 @@ class Memoizer(object):
 
         ignore_list = task['ignore_for_cache']
 
-        logger.debug("Ignoring these kwargs for checkpointing: {}".format(ignore_list))
+        logger.debug("Ignoring these kwargs for checkpointing: %s", ignore_list)
         for k in ignore_list:
-            logger.debug("Ignoring kwarg {}".format(k))
+            logger.debug("Ignoring kwarg %s", k)
             del filtered_kw[k]
 
         if 'outputs' in task['kwargs']:
             outputs = task['kwargs']['outputs']
             del filtered_kw['outputs']
-            t = t + [id_for_memo(outputs, output_ref=True)]   # TODO: use append?
+            t.append(id_for_memo(outputs, output_ref=True))
 
-        t = t + [id_for_memo(filtered_kw)]
-        t = t + [id_for_memo(task['func']),
-                 id_for_memo(task['args'])]
+        t.extend(map(id_for_memo, (filtered_kw, task['func'], task['args'])))
 
         x = b''.join(t)
-        hashedsum = hashlib.md5(x).hexdigest()
-        return hashedsum
+        return hashlib.md5(x).hexdigest()
 
     def check_memo(self, task: TaskRecord) -> Optional[Future[Any]]:
         """Create a hash of the task and its inputs and check the lookup table for this hash.
@@ -289,3 +280,71 @@ class Memoizer(object):
         else:
             logger.debug(f"Storing app cache entry {task['hashsum']} with result from task {task_id}")
         self.memo_lookup_table[task['hashsum']] = r
+
+    def _load_checkpoints(self, checkpointDirs: Sequence[str]) -> Dict[str, Future[Any]]:
+        """Load a checkpoint file into a lookup table.
+
+        The data being loaded from the pickle file mostly contains input
+        attributes of the task: func, args, kwargs, env...
+        To simplify the check of whether the exact task has been completed
+        in the checkpoint, we hash these input params and use it as the key
+        for the memoized lookup table.
+
+        Args:
+            - checkpointDirs (list) : List of filepaths to checkpoints
+              Eg. ['runinfo/001', 'runinfo/002']
+
+        Returns:
+            - memoized_lookup_table (dict)
+        """
+        memo_lookup_table = {}
+
+        for checkpoint_dir in checkpointDirs:
+            logger.info("Loading checkpoints from {}".format(checkpoint_dir))
+            checkpoint_file = os.path.join(checkpoint_dir, 'tasks.pkl')
+            try:
+                with open(checkpoint_file, 'rb') as f:
+                    while True:
+                        try:
+                            data = pickle.load(f)
+                            # Copy and hash only the input attributes
+                            memo_fu: Future = Future()
+                            assert data['exception'] is None
+                            memo_fu.set_result(data['result'])
+                            memo_lookup_table[data['hash']] = memo_fu
+
+                        except EOFError:
+                            # Done with the checkpoint file
+                            break
+            except FileNotFoundError:
+                reason = "Checkpoint file was not found: {}".format(
+                    checkpoint_file)
+                logger.error(reason)
+                raise BadCheckpoint(reason)
+            except Exception:
+                reason = "Failed to load checkpoint: {}".format(
+                    checkpoint_file)
+                logger.error(reason)
+                raise BadCheckpoint(reason)
+
+            logger.info("Completed loading checkpoint: {0} with {1} tasks".format(checkpoint_file,
+                                                                                  len(memo_lookup_table.keys())))
+        return memo_lookup_table
+
+    @typeguard.typechecked
+    def load_checkpoints(self, checkpointDirs: Optional[Sequence[str]]) -> Dict[str, Future]:
+        """Load checkpoints from the checkpoint files into a dictionary.
+
+        The results are used to pre-populate the memoizer's lookup_table
+
+        Kwargs:
+             - checkpointDirs (list) : List of run folder to use as checkpoints
+               Eg. ['runinfo/001', 'runinfo/002']
+
+        Returns:
+             - dict containing, hashed -> future mappings
+        """
+        if checkpointDirs:
+            return self._load_checkpoints(checkpointDirs)
+        else:
+            return {}
