@@ -1,18 +1,34 @@
 import inspect
 import logging
 import os
+import re
 import shlex
 import subprocess
 import threading
 import time
-import typeguard
 from contextlib import contextmanager
-from typing import List, Tuple, Union, Generator, IO, AnyStr, Dict
+from types import TracebackType
+from typing import (
+    IO,
+    Any,
+    AnyStr,
+    Callable,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
+
+import typeguard
+from typing_extensions import Type
 
 import parsl
+from parsl.app.errors import BadStdStreamFile
 from parsl.version import VERSION
 
-
 try:
     import setproctitle as setproctitle_module
 except ImportError:
@@ -44,7 +60,7 @@ def get_version() -> str:
 
 
 @typeguard.typechecked
-def get_all_checkpoints(rundir: str = "runinfo") -> List[str]:
+def get_all_checkpoints(rundir: str = "runinfo") -> Sequence[str]:
     """Finds the checkpoints from all runs in the rundir.
 
     Kwargs:
@@ -55,7 +71,7 @@ def get_all_checkpoints(rundir: str = "runinfo") -> List[str]:
 
     """
 
-    if(not os.path.isdir(rundir)):
+    if not os.path.isdir(rundir):
         return []
 
     dirs = sorted(os.listdir(rundir))
@@ -73,7 +89,7 @@ def get_all_checkpoints(rundir: str = "runinfo") -> List[str]:
 
 
 @typeguard.typechecked
-def get_last_checkpoint(rundir: str = "runinfo") -> List[str]:
+def get_last_checkpoint(rundir: str = "runinfo") -> Sequence[str]:
     """Finds the checkpoint from the last run, if one exists.
 
     Note that checkpoints are incremental, and this helper will not find
@@ -99,38 +115,41 @@ def get_last_checkpoint(rundir: str = "runinfo") -> List[str]:
     last_runid = dirs[-1]
     last_checkpoint = os.path.abspath(f'{rundir}/{last_runid}/checkpoint')
 
-    if(not(os.path.isdir(last_checkpoint))):
+    if not os.path.isdir(last_checkpoint):
         return []
 
     return [last_checkpoint]
 
 
-def get_std_fname_mode(fdname: str, stdfspec: Union[str, Tuple[str, str]]) -> Tuple[str, str]:
+@typeguard.typechecked
+def get_std_fname_mode(
+    fdname: str,
+    stdfspec: Union[os.PathLike, str, Tuple[str, str], Tuple[os.PathLike, str]]
+) -> Tuple[str, str]:
     import parsl.app.errors as pe
-    if stdfspec is None:
-        return None, None
-    elif isinstance(stdfspec, str):
+    if isinstance(stdfspec, (str, os.PathLike)):
         fname = stdfspec
         mode = 'a+'
     elif isinstance(stdfspec, tuple):
         if len(stdfspec) != 2:
             msg = (f"std descriptor {fdname} has incorrect tuple length "
                    f"{len(stdfspec)}")
-            raise pe.BadStdStreamFile(msg, TypeError('Bad Tuple Length'))
+            raise pe.BadStdStreamFile(msg)
         fname, mode = stdfspec
-        if not isinstance(fname, str) or not isinstance(mode, str):
-            msg = (f"std descriptor {fdname} has unexpected type "
-                   f"{type(stdfspec)}")
-            raise pe.BadStdStreamFile(msg, TypeError('Bad Tuple Type'))
+
+    path = os.fspath(fname)
+
+    if isinstance(path, str):
+        return path, mode
+    elif isinstance(path, bytes):
+        return path.decode(), mode
     else:
-        msg = f"std descriptor {fdname} has unexpected type {type(stdfspec)}"
-        raise pe.BadStdStreamFile(msg, TypeError('Bad Tuple Type'))
-    return fname, mode
+        raise BadStdStreamFile(f"fname has invalid type {type(path)}")
 
 
 @contextmanager
 def wait_for_file(path: str, seconds: int = 10) -> Generator[None, None, None]:
-    for i in range(0, int(seconds * 100)):
+    for _ in range(0, int(seconds * 100)):
         time.sleep(seconds / 100.)
         if os.path.exists(path):
             break
@@ -167,7 +186,7 @@ def wtime_to_minutes(time_string: str) -> int:
     return total_mins
 
 
-class RepresentationMixin(object):
+class RepresentationMixin:
     """A mixin class for adding a __repr__ method.
 
     The __repr__ method will return a string equivalent to the code used to instantiate
@@ -193,8 +212,10 @@ class RepresentationMixin(object):
     """
     __max_width__ = 80
 
+    _validate_repr = False
+
     def __repr__(self) -> str:
-        init = self.__init__  # type: ignore
+        init = self.__init__  # type: ignore[misc]
 
         # This test looks for a single layer of wrapping performed by
         # functools.update_wrapper, commonly used in decorators. This will
@@ -213,18 +234,21 @@ class RepresentationMixin(object):
         else:
             defaults = {}
 
-        for arg in argspec.args[1:]:
-            if not hasattr(self, arg):
-                template = (f'class {self.__class__.__name__} uses {arg} in the'
-                            f' constructor, but does not define it as an '
-                            f'attribute')
-                raise AttributeError(template)
+        if self._validate_repr:
+            for arg in argspec.args[1:]:
+                if not hasattr(self, arg):
+                    template = (f'class {self.__class__.__name__} uses {arg} in the'
+                                f' constructor, but does not define it as an '
+                                f'attribute')
+                    raise AttributeError(template)
+
+        default = "<unrecorded>"
 
         if len(defaults) != 0:
-            args = [getattr(self, a) for a in argspec.args[1:-len(defaults)]]
+            args = [getattr(self, a, default) for a in argspec.args[1:-len(defaults)]]
         else:
-            args = [getattr(self, a) for a in argspec.args[1:]]
-        kwargs = {key: getattr(self, key) for key in defaults}
+            args = [getattr(self, a, default) for a in argspec.args[1:]]
+        kwargs = {key: getattr(self, key, default) for key in defaults}
 
         def assemble_multiline(args: List[str], kwargs: Dict[str, object]) -> str:
             def indent(text: str) -> str:
@@ -255,11 +279,11 @@ class AtomicIDCounter:
     """A class to allocate counter-style IDs, in a thread-safe way.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.count = 0
         self.lock = threading.Lock()
 
-    def get_id(self):
+    def get_id(self) -> int:
         with self.lock:
             new_id = self.count
             self.count += 1
@@ -271,3 +295,203 @@ def setproctitle(title: str) -> None:
         setproctitle_module.setproctitle(title)
     else:
         logger.warn(f"setproctitle not enabled for process {title}")
+
+
+class Timer:
+    """This class will make a callback periodically, with a period
+    specified by the interval parameter.
+
+    This is based on the following logic :
+
+    .. code-block:: none
+
+
+        BEGIN (INTERVAL, THRESHOLD, callback) :
+            start = current_time()
+
+            while (current_time()-start < INTERVAL) :
+                 wait()
+                 break
+
+            callback()
+
+    """
+
+    def __init__(self, callback: Callable, *args: Any, interval: Union[float, int] = 5, name: Optional[str] = None) -> None:
+        """Initialize the Timer object.
+        We start the timer thread here
+
+        KWargs:
+             - interval (int or float) : number of seconds between callback events
+             - name (str) : a base name to use when naming the started thread
+        """
+
+        self.interval = max(0, interval)
+        self.cb_args = args
+        self.callback = callback
+
+        self._kill_event = threading.Event()
+        tname = f"Timer-Thread-{id(self)}"
+        if name:
+            tname = f"{name}-{tname}"
+        self._thread = threading.Thread(
+            target=self._wake_up_timer, name=tname, daemon=True
+        )
+        self._thread.start()
+
+    def _wake_up_timer(self) -> None:
+        self.make_callback()
+
+        while not self._kill_event.wait(self.interval):
+            self.make_callback()
+
+    def make_callback(self) -> None:
+        """Makes the callback and resets the timer.
+        """
+        try:
+            self.callback(*self.cb_args)
+        except Exception:
+            logger.error("Callback threw an exception - logging and proceeding anyway", exc_info=True)
+
+    def close(self, timeout: Optional[float] = None) -> None:
+        """Merge the threads and terminate.
+        """
+        self._kill_event.set()
+        self._thread.join(timeout=timeout)
+
+
+class AutoCancelTimer(threading.Timer):
+    """
+    Extend threading.Timer for use as a context manager
+
+    Example:
+
+        with AutoCancelTimer(delay, your_callback):
+            some_func()
+
+    If `some_func()` returns before the delay is up, the timer will
+    be cancelled.
+    """
+    def __enter__(self) -> "AutoCancelTimer":
+        self.start()
+        return self
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc_val: Optional[BaseException],
+        exc_tb: Optional[TracebackType]
+    ) -> None:
+        self.cancel()
+
+
+def sanitize_dns_label_rfc1123(raw_string: str) -> str:
+    """Convert input string to a valid RFC 1123 DNS label.
+
+    Parameters
+    ----------
+    raw_string : str
+        String to sanitize.
+
+    Returns
+    -------
+    str
+        Sanitized string.
+
+    Raises
+    ------
+    ValueError
+        If the string is empty after sanitization.
+    """
+    # Convert to lowercase and replace non-alphanumeric characters with hyphen
+    sanitized = re.sub(r'[^a-z0-9]', '-', raw_string.lower())
+
+    # Remove consecutive hyphens
+    sanitized = re.sub(r'-+', '-', sanitized)
+
+    # DNS label cannot exceed 63 characters
+    sanitized = sanitized[:63]
+
+    # Strip after trimming to avoid trailing hyphens
+    sanitized = sanitized.strip("-")
+
+    if not sanitized:
+        raise ValueError(f"Sanitized DNS label is empty for input '{raw_string}'")
+
+    return sanitized
+
+
+def sanitize_dns_subdomain_rfc1123(raw_string: str) -> str:
+    """Convert input string to a valid RFC 1123 DNS subdomain.
+
+    Parameters
+    ----------
+    raw_string : str
+        String to sanitize.
+
+    Returns
+    -------
+    str
+        Sanitized string.
+
+    Raises
+    ------
+    ValueError
+        If the string is empty after sanitization.
+    """
+    segments = raw_string.split('.')
+
+    sanitized_segments = []
+    for segment in segments:
+        if not segment:
+            continue
+        sanitized_segment = sanitize_dns_label_rfc1123(segment)
+        sanitized_segments.append(sanitized_segment)
+
+    sanitized = '.'.join(sanitized_segments)
+
+    # DNS subdomain cannot exceed 253 characters
+    sanitized = sanitized[:253]
+
+    # Strip after trimming to avoid trailing dots or hyphens
+    sanitized = sanitized.strip(".-")
+
+    if not sanitized:
+        raise ValueError(f"Sanitized DNS subdomain is empty for input '{raw_string}'")
+
+    return sanitized
+
+
+def execute_wait(cmd: str, walltime: Optional[int] = None) -> Tuple[int, str, str]:
+    ''' Synchronously execute a commandline string on the shell.
+
+    Args:
+        - cmd (string) : Commandline string to execute
+        - walltime (int) : walltime in seconds
+
+    Returns:
+        - retcode : Return code from the execution
+        - stdout  : stdout string
+        - stderr  : stderr string
+    '''
+    try:
+        logger.debug("Creating process with command '%s'", cmd)
+        proc = subprocess.Popen(
+            cmd,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            shell=True,
+            preexec_fn=os.setpgrp
+        )
+        logger.debug("Created process with pid %s. Performing communicate", proc.pid)
+        (stdout, stderr) = proc.communicate(timeout=walltime)
+        retcode = proc.returncode
+        logger.debug("Process %s returned %s", proc.pid, proc.returncode)
+
+    except Exception:
+        logger.exception(f"Execution of command failed:\n{cmd}")
+        raise
+    else:
+        logger.debug("Execution of command in process %s completed normally", proc.pid)
+
+    return (retcode, stdout.decode("utf-8"), stderr.decode("utf-8"))
